/*
   Camera Rotations Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create a basic camera
   system that can be used in DirectX.
*/

// Demo: Camera Rotations.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>
#include"CCamera.h"


// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1)


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObjects();
void SetCamera();
void SetPerpective();
void RenderScene();
void ShutdownDirect3D();
void GetInput();


// Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;
LPDIRECT3DVERTEXBUFFER9 Object_Buffer = NULL;
LPDIRECT3DVERTEXBUFFER9 Floor_Buffer = NULL;
LPDIRECT3DTEXTURE9 Texture = NULL;


// Camera object used to move around the 3D scene.
CCamera Camera;


// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   DWORD color;
   FLOAT tu, tv;
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;
               }
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Camera Rotations Demo - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;


   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;
   
   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg); // Translate any messages.
               DispatchMessage(&msg);  // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }


    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		{
		   return false;
	   }

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.


   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }


   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // SetRenderState() will set a parameter for the rendering.  Here we turn off lighting.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);

   // Disable culling.  This will render the front and back of the triangles.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

   // Enable depth testing.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);

   // Here we set up the object this demo will render (for example a triangle).
   if(!InitializeObjects())
      return false;

   // Here we set the demos perpective.
   SetPerpective();

   // Set the image states to get a good quality image.
	D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);


   // Here we set the camera's default position.  SetCamera requires the x, y, z position
   // of the camera, next it needs the x, y, z position the camera is looking at, and last
   // it needs the direction for which way is up (which is a positive number on the y axis).
   Camera.SetCamera(0.0f, 0.0f, -2.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);

   return true;
}


bool InitializeObjects()
{
   // Triangle data for its size and color.
	Vertex Triangle_Data[3] =
   {
		{-0.1f, -0.1f, 0.0f, D3DCOLOR_XRGB(255,255,0)},
      {0.1f, -0.1f, 0.0f, D3DCOLOR_XRGB(255,255,0)},
	   {0.0f, 0.1f, 0.0f, D3DCOLOR_XRGB(0,0,255)}
	};

   // Create the vertex buffer that will hold the triangle.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Triangle_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Object_Buffer, NULL)))
      {
         return false;
      }
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Object_Buffer->Lock(0, sizeof(Triangle_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the triangle's data into the vertex buffer.
   memcpy(Vertices, Triangle_Data, sizeof(Triangle_Data));

   // Unlock when your done coping data into the buffer.
   Object_Buffer->Unlock();

   
   // Create the ground.
   Vertex Floor_Data[] =
   {
      {-4.0f, -0.2f,-4.0f, D3DCOLOR_XRGB(255,255,255), 0.0f, 1.0f},
	   {-4.0f, -0.2f, 4.0f, D3DCOLOR_XRGB(255,255,255), 0.0f, 0.0f},
	   { 4.0f, -0.2f,-4.0f, D3DCOLOR_XRGB(255,255,255), 1.0f, 1.0f},
	   { 4.0f, -0.2f, 4.0f, D3DCOLOR_XRGB(255,255,255), 1.0f, 0.0f}
	};


   // Create the vertex buffer that will hold the triangle.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Floor_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Floor_Buffer, NULL)))
      {
         return false;
      }
	
   // Lock the buffer we can write to it.
   if(FAILED(Floor_Buffer->Lock(0, sizeof(Floor_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the triangle's data into the vertex buffer.
   memcpy(Vertices, Floor_Data, sizeof(Floor_Data));

   // Unlock when your done coping data into the buffer.
   Floor_Buffer->Unlock();


   // Here we load the texture into the texture object.
   if(D3DXCreateTextureFromFile(D3D_Device, "ground.bmp", &Texture) != D3D_OK)
      return false;

   return true;
}


void SetCamera()
{
   // This will be used to set the view ("camera") matrix.
   D3DXMATRIXA16 View;

   // Get the input from the user so we can move the camera either foward or back.
   GetInput();

   // Eye is the position, LookAt is the point the camera is looking at, and Up
   // just tells Direct3D which way is up (positive 1 and above is...well up).

   // Now we set the camera data into the 3 D3DXVECTOR3 objects we will send to create the
   // view matrix so we can use it.  You can use the arrow keys to move the camera.
   D3DXVECTOR3 Eye(Camera.xPos, Camera.yPos, Camera.zPos);        // Position of camera.
   D3DXVECTOR3 LookAt(Camera.xView, Camera.yView, Camera.zView);  // Point camera is looking at.
   D3DXVECTOR3 Up(Camera.xUp, Camera.yUp, Camera.zUp);            // Which way is up.

   // Set all the values this matrix will need to know (position, look at point, etc).
   D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);

   // Move the "camera" to the position and values we set above.
   D3D_Device->SetTransform(D3DTS_VIEW, &View);
}


void SetPerpective()
{
   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the info into the matrix.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, 640/480, 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void RenderScene()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(127, 127, 255), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Set the view matrix (camera position).
      SetCamera();

      D3D_Device->SetTexture(0, NULL);

	   // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Object_Buffer, 0, sizeof(Vertex));

      // Set the vertex stream that was defined at the top.
      D3D_Device->SetFVF(D3DFVF_D3DVertex);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 1);


      // Draw the floor now.

      // Apply texture image.
      D3D_Device->SetTexture(0, Texture);
      D3D_Device->SetStreamSource(0, Floor_Buffer, 0, sizeof(Vertex));
      D3D_Device->SetFVF(D3DFVF_D3DVertex);
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
    
   // We are done drawing.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Object_Buffer != NULL)
      {
	      Object_Buffer->Release();
         Object_Buffer = NULL;
      }

   if(Floor_Buffer != NULL)
      {
	      Floor_Buffer->Release();
         Floor_Buffer = NULL;
      }
}


void GetInput()
{
   // Here we check for each input that we care about.
   if(GetKeyState(VK_UP) & 0x80)
      Camera.MoveCamera(UP);

   if(GetKeyState(VK_DOWN) & 0x80)
      Camera.MoveCamera(DOWN);
}


// Recap:

// Its not much but its a start.  Next we'll add camera rotations.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com